#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "wisck.h"

static const char * optstring = "LROTVCcbBnNvs:S:o:";

static void usage() 
{
    fprintf(stderr,"usage:\n");
    fprintf(stderr,"wisck -LROTVC -c -bB -nN -s vsep -S csep -v[v] -o outfile infile1 infile2 ...\n");
    exit(EX_USAGE);
}

int main(int argc, char *argv[])
{
    uint32_t flags;
    char *ofn;
    int infd, outfd;
    unsigned char vsep, csep;
    int ch;
    
    /*by default sielently output everything*/
    flags  = WISCK_OLIN | WISCK_OREC | WISCK_OOBJ;
    flags |= WISCK_ONAM | WISCK_OVAL | WISCK_OCOM;
    flags |= WISCK_UCSE | WISCK_UVSE;
    ofn = NULL;
    vsep = csep = '\n';
    infd = 0;
    outfd = 1;
    ch = 0;

    while( -1 != (ch = getopt(argc, argv, optstring))) {
        switch(ch) {
            case '?':
                usage();
                break;
            case 'L':
                flags &= ~WISCK_OLIN;
                break;
            case 'R':
                flags &= ~WISCK_OREC;
                break;
            case 'O':
                flags &= ~WISCK_OOBJ;
                break;
            case 'T':
                flags &= ~WISCK_ONAM;
                break;
            case 'V':
                flags &= ~WISCK_OVAL;
                break;
            case 'C':
                flags &= ~WISCK_OCOM;
                break;
            case 'c':
                flags |= WISCK_SCOM;
                break;
            case 'b':
                flags |= WISCK_SBBV;
                break;
            case 'B':
                flags |= WISCK_SBBC;
                break;
            case 'n':
                flags &= ~WISCK_UVSE;
                break;
            case 'N':
                flags &= ~WISCK_UCSE;
                break;
            case 's':
                if (strlen(optarg) > 1)
                   usage(); 
                vsep = optarg[0];
                break;
            case 'S':
                if (strlen(optarg) > 1)
                    usage();
                csep = optarg[0];
                break;
            case 'v':
                BLIN_VERBOSE(flags);
                break;
            case 'o':
                ofn = strdup(optarg); 
                break;
        }
    }
    argc -= optind;
    argv += optind;
    if (ofn != NULL) {
        if ((outfd = open(ofn, O_RDWR|O_CREAT|O_TRUNC, 0666)) < 0) {
            ifBLIN_QV1(flags)
                warn("can't open output file");
            free(ofn);
            return EX_CANTCREAT;
        }
        free(ofn);
    }
    else {
        outfd = fileno(stdout);
    }
    if (argc == 0) {
        if (wisck_convert(0, outfd, flags, vsep, csep) < 0) {
            if (outfd != fileno(stdout)) 
                close(outfd);
            return EX_SOFTWARE;        
        }
    } 
    else {
        while (argc-- > 0) {
            if ((infd = open(*argv, O_RDONLY)) < 0) {
                ifBLIN_QV1(flags)
                    warn("can't open input file %s", *argv);
                if (outfd != fileno(stdout)) 
                    close(outfd);
                return EX_NOINPUT;
            }
            argv++;
            if (wisck_convert(infd, outfd, flags, vsep, csep) < 0) {
                if (outfd != fileno(stdout)) 
                    close(outfd);
                return EX_SOFTWARE;        
            }
            close(infd);
        }
    }
    if (outfd != fileno(stdout))
        close(outfd);
    return 0;
}

