/* Who Is SuCK library */

/*
 * Flags of parse procedure
 */
#define WISCK_SCOM  0x0001    /*skip comments*/
#define WISCK_UVSE  0x0002    /*use value parts separating symbol (vsep)*/
#define WISCK_UCSE  0x0004    /*use comment parts separating symbol (csep) */
#define WISCK_SBBV  0x0008    /*skip blank characters before value part*/
#define WISCK_SBBC  0x0010    /*skip blank characters before comment part*/

/* 
 * Flags of output procedure
 */
#define WISCK_OLIN  0x0080    /*output line number*/
#define WISCK_OREC  0x0100    /*output record number*/
#define WISCK_OOBJ  0x0200    /*output object number*/
#define WISCK_ONAM  0x0400    /*output record name*/
#define WISCK_OVAL  0x0800    /*output record value*/
#define WISCK_OCOM  0x1000    /*output record comments*/

typedef struct wisck_descriptor {
    uint32_t flags;                 /*flags. The highest byte is BLIN flags*/               
    u_char vsep;             /*value part separator (usualy \n)*/
    u_char csep;             /*comment part separator (usualy \n)*/
    mife_descriptor *mife;         /*window*/
    mife_aout * aout;               /*async output structure*/
    off_t ind;                      /*position of symbol to be read in the stream*/
    u_char *buf;             /*prepared buffer for parsing*/
    uint32_t line;                  /*line of input file number*/
    uint32_t rec;                   /*input record number (record means one name - value property)*/
    uint32_t obj;                   /*object number*/
} wisck_descriptor;

typedef struct wisck_rec {
    off_t offset;       /*offset of current record in input file*/
    uint32_t line;      /*record begining line number*/
    uint32_t rec;       /*recored number*/
    uint32_t obj;       /*object number*/
    char *name;         /*pointer to name*/
    char *value;        /*pointer to value*/
    char *comm;         /*pointer to comment*/
    size_t nlen;        /*name length*/
    size_t vlen;        /*value length*/
    size_t clen;        /*comment length*/
    u_char *tail;/*pointer to unparsed part of buffer*/
} wisck_rec;

/*
 * Classes of symbols
 */
enum sclass {
    WISCK_ALNUM,   /*letter-digit*/
    WISCK_COLON,   /*':'*/
    WISCK_SPACE,   /*' ' or '\t'*/
    WISCK_NEWLINE,   /*'\n'*/
    WISCK_HASH,   /*'#'*/
    WISCK_PLUS,   /*'+'*/
    WISCK_ANOTHER,   /*another*/
    WISCK_EOF    /*truth EOF*/
};    
/*
 * Create new wisck descriptor initialized with fd file descriptor
 */
wisck_descriptor *
wisck_init(int infd, int outfd, u_int32_t flags, u_char vsep, u_char csep); 

/*
 * Prepare bufer for parsing
 */
int
wisck_prepare(wisck_descriptor *wisck, struct wisck_rec *rec);

/*
 * Parse one record
 */
int
wisck_parse(wisck_descriptor *wisck, struct wisck_rec *rec);
        
/*
 * Close wisck descriptor, releasing allocated memory
 */
int 
wisck_close(wisck_descriptor *wisck);

/*
 * Performs complete file converting. facade-function
 */
int
wisck_convert(int infd, int outfd, uint32_t flags, u_char vsep, u_char csep);

