#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <sys/types.h>
#include <babolo/BLINflag.h>
#include <sys/uio.h>
#include <mife.h>
#include "wisck.h"

/*
 * object separator in input file
 */
#define OSEP "\n\n"

/*
 * this char will separate fields of outpur
 */
#define RSEP '\0'

/* 
 * actions 
 * 16 most significant bits
 */
#define GET_ACTION(x) ((x) & 0xFFFF0000)

#define Il 0x00000100   /*increase line number (increased on each \n)*/
#define Io 0x00000200   /*increase object number*/
#define Ir 0x00000400   /*increase record number (increased when new record detected, whew first symbol of name detected or standalone comment found)*/

#define Ao 0x00000800   /*append ordinary character of field name or field value*/
#define Ac 0x00001000   /*append character of comment (may disabled by flag)*/

#define Vb 0x00002000   /*append blank character before value (may be disabled by flag)*/
#define Cb 0x00004000   /*append blank character before comment (may be disabled by flag)*/

#define Vs 0x00008000   /*append value part separator*/
#define Cs 0x00010000   /*append comment part separator (or record separator after first value part) */
#define Rs 0x00020000   /*append zero unconditionally*/
#define Rc 0x00040000   /*append zero after first value part*/
#define Cc 0x00080000   /*append zero after comment (if it was)*/

#define Np 0x00100000   /*set name pointer*/
#define Vp 0x00200000   /*set value pointer*/
#define Cp 0x00400000   /*set comment pointer*/
#define Ap 0x00800000   /*set additional pointer*/

#define Nx 0x01000000   /*increase reading ind*/
#define Sr 0x02000000   /*sort*/
#define Nu 0x04000000   /*return NULL*/

#define GET_STATE(x) ((x) & 0x000000FF)

/*
 * States of main FSM - for single line records
 */
enum mstate {   
    Sin,    /*initial state - waiting for first symbol of next object*/
    Snm,    /*reading name*/
    Sb1,    /*droping spaces and tabs before first symbol of value*/
    Svl,    /*reading value*/
    Sb3,    /*droping blanks before comment (if enabled)*/
    Sco,    /*reading comments*/
    Snl,    /*detecting multiline value*/
    Ser,    /*failed*/
    Sfi     /*finished*/
};  

enum _sclass {
    S_CH = WISCK_ALNUM,   /*letter-digit*/
    S_DD = WISCK_COLON,   /*':'*/
    S_BL = WISCK_SPACE,   /*' ' or '\t'*/
    S_NL = WISCK_NEWLINE,   /*'\n'*/
    S_HA = WISCK_HASH,   /*'#'*/
    S_PL = WISCK_PLUS,   /*'+'*/
    S_AN = WISCK_ANOTHER,   /*another*/
    S_EN = WISCK_EOF   /*truth EOF*/
};    

/*classificator*/
static const enum _sclass res[256] = 
{S_EN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_BL, S_NL, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_BL, S_AN, S_AN, S_HA,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_PL,   S_AN, S_CH, S_AN, S_AN
,S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_DD, S_AN,   S_AN, S_AN, S_AN, S_AN

,S_AN, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH
,S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH
,S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_CH,   S_CH, S_CH, S_CH, S_AN,   S_AN, S_AN, S_AN, S_AN

,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN

,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
,S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN,   S_AN, S_AN, S_AN, S_AN
};

static const uint32_t fsm[Ser+1][S_EN+1] =
/*       S_CH          S_DD          S_BL          S_NL                S_HA             S_PL          S_AN       S_EN*/
{/*Sin*/{Snm|Np|Ao|Nx, Sb1|Vp|Nx,    Ser|Nx,       Sin|Il|Nx,          Sb3|Cp|Nx,       Ser|Nx,       Ser|Nx,    Sfi|Nu}
,/*Snm*/{Snm|Ao|Nx,    Sb1|Rs|Vp|Nx, Ser|Nx,       Ser|Il|Nx,          Ser|Nx,          Ser|Nx,       Ser|Nx,    Ser   }
,/*Sb1*/{Svl|Ao|Nx,    Svl|Ao|Nx,    Sb1|Vb|Nx,    Snl|Il|Rc|Sr|Nx,    Sb3|Cp|Cs|Sr|Nx, Svl|Ao|Nx,    Svl|Ao|Nx, Sfi|Rs}
,/*Svl*/{Svl|Ao|Nx,    Svl|Ao|Nx,    Svl|Ao|Nx,    Snl|Il|Rc|Sr|Nx,    Sb3|Cp|Cs|Sr|Nx, Svl|Ao|Nx,    Svl|Ao|Nx, Sfi|Rs}
,/*Sb3*/{Sco|Ac|Nx,    Sco|Ac|Nx,    Sb3|Cb|Nx,    Snl|Il|Nx,          Sco|Ac|Nx,       Sco|Ac|Nx,    Sco|Ac|Nx, Sfi|Rs}
,/*Sco*/{Sco|Ac|Nx,    Sco|Ac|Nx,    Sco|Ac|Nx,    Snl|Il|Nx,          Sco|Ac|Nx,       Sco|Ac|Nx,    Sco|Ac|Nx, Sfi|Rs}
,/*Snl*/{Sfi|Cc|Ir,    Sfi|Cc|Ir,    Sb1|Ap|Vs|Nx, Sfi|Cc|Il|Ir|Io|Nx, Sfi|Cc|Ir,       Sb1|Ap|Vs|Nx, Ser|Ir|Nx, Sfi|Rs}
,/*Ser*/{Ser|Nx,       Ser|Nx,       Ser|Nx,       Sfi|Il|Nx,          Ser|Nx,          Ser|Nx,       Ser|Nx,    Sfi   }
};

/*==================== debug =====================*/
static const char *sclass_inv[] = {"S_CH","S_DD","S_BL","S_NL","S_HA","S_PL","S_AN","S_EN"};

static const char *state_inv[] = {"Sin","Snm","Sb1","Svl","Sb3","Sco","Snl","Ser","Sfi"};
/*==================== debug =====================*/

/*==================== sort ======================*/
/*
 * inverse array
 * p - pointer
 * s - index of begening of block to be inverted
 * e - index of first symbol after block to be inverted
 */
static void 
inverse(u_char * p, off_t s, off_t e) 
{
    char tmp;
    off_t i;

    for (i=0; i< (e-s)/2; i++) {
        tmp = p[s + i];
        p[s + i] = p[e - 1 - i];
        p[e - 1 - i] = tmp;
    }
}

/*
 * Performs cyclic left shift of buffer pointed by p
 * with length head + tail bytes
 * head - length of head part of buffer that must moved
 *        at the end
 * tail - length og tail of buffer
 * NOTE : tail part of buffer must follow head part wihout a gap
 */
static void 
lshift(u_char * p, ssize_t head, ssize_t tail) 
{
    inverse(p, 0, head);
    inverse(p, head, head + tail);
    inverse(p, 0, head + tail);
}
/*==================== sort =======================*/


/*
 * Create new wisck descriptor initialized with fd file descriptor
 */
struct wisck_descriptor *
wisck_init(int infd, int outfd, u_int32_t flags, unsigned char vsep, unsigned char csep) 
{
    struct wisck_descriptor * wisck;
    uint32_t mife_flags;

    if (NULL == (wisck = (struct wisck_descriptor *)malloc(sizeof(struct wisck_descriptor)))) {
        errno = ENOMEM;
        return NULL;
    }
    memset(wisck,0,sizeof(struct wisck_descriptor));
    wisck->flags = flags;
    wisck->vsep = vsep;
    wisck->csep = csep;
    mife_flags = MIFE_FULL|MIFE_SLID;
    wisck->mife = mife_opef(mife_flags,infd);
    if (-1 == mife_eolinit(wisck->mife, OSEP)) {
        free(wisck);
        ifBLIN_QV1(flags)
            warn("mife_eolinit failed");
        return NULL;
    }
    if (-1 == mife_read(wisck->mife, 0, 0)) {
        free(wisck);
        ifBLIN_QV1(flags)
            warn("initial mife_read failed");
        return NULL;
    }
    if (NULL == (wisck->aout = mife_aopen(mife_flags, outfd))) {
        free(wisck);
        ifBLIN_QV1(flags)
            warn("mife_aopen failed");
        return NULL;
    }
    wisck->ind = 0;
    wisck->line = 1;
    wisck->obj = 1;
    wisck->rec = 1;
    return wisck;
}

/*
 * Prepare bufer for parsing
 */
int
wisck_prepare(struct wisck_descriptor *wisck, struct wisck_rec *rec) 
{
    if (-1 == mife_read(wisck->mife, 0, wisck->ind)) {
        ifBLIN_QV1(wisck->flags)
            warn("mife read error");
        return -1;
    }
    if (NULL == (wisck->buf = mife_get(wisck->mife,wisck->ind))) {
        ifBLIN_QV1(wisck->flags)
            warn("mife get error");
        return -1;
    }  
    return 0;
}

/*
 * Parse one record
 */
int
wisck_parse(struct wisck_descriptor *wisck, struct wisck_rec *rec) 
{
    /*declarations*/
    unsigned char   *rd,     /*local read index*/
                    *hwm,    /*high water mark*/
                    *np,     /*name index*/
                    *vp,     /*value index*/
                    *cp,     /*comment index*/
                    *ap;     /*additional index*/
    char    isv0;   /*is '\0' set for values*/
    enum mstate state;
    enum _sclass sym;
    uint32_t    action;
    unsigned char   c;

    state = Sin;
    np = vp = cp = ap = NULL;
    isv0 = 0;
    hwm = rd = wisck->buf;
    rec->offset = wisck->ind;
    rec->line = wisck->line;
    rec->rec = wisck->rec;
    rec->obj = wisck->obj;
    rec->name = rec->value = rec->comm = NULL;
    rec->nlen = rec->vlen = rec->clen = 0;
    rec->tail = NULL;

    while (state != Sfi) {
        c = (unsigned char) *rd;
        if (c == '\0' && !(wisck->mife->flag & MIFE_EOFL)) {
            ifBLIN_QV1(wisck->flags)
                fprintf(stderr,"unexpected zero in buffer\n");
            return -1;
        }
        sym = res[c];
        action = fsm[state][sym];
        ifBLIN_QV3(wisck->flags)
            fprintf(stderr,"%s( %s [%c] )->%s\n", state_inv[state], sclass_inv[sym], c, state_inv[GET_STATE(action)]);
        ifBLIN_QV4(wisck->flags)
            fprintf(stderr,"before rd %p hwm %p\n", rd, hwm);
        state = GET_STATE(action);
        /************/
        if (action & Il) wisck->line++;
        if (action & Ir) wisck->rec++;
        if (action & Io) wisck->obj++;
        /************/
        if (action & Sr) {
            if (cp != NULL) {
                lshift(cp - 1, 1 + ap - cp, hwm - ap);
                cp += hwm - ap;
            }
            if (cp == NULL && ap != NULL) {
                lshift(ap - 1, 1, hwm - ap);
            }
        }
        /************/
        if (action & Np) {
            np = hwm;
        }
        if (action & Vp) {
            vp = hwm + 1;
        }
        if (action & Cs) {
            if (!isv0) {
                *hwm++ = RSEP;
                isv0 = 1;
            } else {
                if ((cp != NULL) && (wisck->flags & WISCK_UCSE))
                *hwm++ = wisck->csep;
            }
        }
        if (action & Cp) {
            if (cp == NULL) {
                if (vp != NULL &&!isv0)
                    cp = hwm + 1; /*first comment part after first value part. # should be replaced with \0*/
                else
                    cp = hwm;     /*value is not set(only comment in line) or \0 for value set before*/
            }
        }
        if (action & Ap) {
            ap = hwm;
        }
        /************/
        if (action & Vs) {
                if (wisck->flags & WISCK_UVSE)
                    *hwm++ = wisck->vsep;
        }
        if (action & Rs) *hwm++ = RSEP;
        if (action & Rc) {
            if (!isv0) {
                *hwm++ = RSEP;
                isv0 = 1;
            }
        }
        if (action & Cc) {
            if (cp != NULL)
                *hwm++ = RSEP;
        }
        if (action & Ao) *hwm++ = c;
        if (action & Ac) {
            if (!(wisck->flags & WISCK_SCOM)) 
                *hwm++ = c; 
        }
        if (action & Vb) {
            if (!(wisck->flags & WISCK_SBBV))
                *hwm++ = c;
        }
        if (action & Cb) {
            if (!(wisck->flags & WISCK_SBBC))
                *hwm++ = c;
        }
        /************/
        if (action & Nu) {
            wisck->ind += rd - wisck->buf;
            return 0;
        }
        if (action & Nx) rd++;

        ifBLIN_QV4(wisck->flags)
            fprintf(stderr, "after rd %p hwm %p\n", rd, hwm);
        if (hwm > rd) {
            ifBLIN_QV1 (wisck->flags)
                fprintf(stderr, "internal logic error! hwm %p > rd %p\n", hwm, rd);
            return -1;
        }
    }
    wisck->ind += rd - wisck->buf;
    rec->name = np;
    rec->value = vp;
    rec->comm = cp;
    if (np) 
        rec->nlen = vp - np - 1;
    if (vp)
        rec->vlen = (cp ? cp : hwm) - vp - 1;
    if (cp)
        rec->clen = hwm - cp - 1;
    rec->tail = rd;
    return 1; 
}

/*
 * outputs zero ended string inserting
 * escape symbols
 * TODO: check errors
 */
static void
wisck_escape_output(mife_aout * aout, char *str) 
{
    unsigned char c;

    while( '\0' != (c = *str++))
        switch(c) {
            case '\b':
                mife_aput(aout,'\\');
                mife_aput(aout,'b');
                break;
            case '\n':
                mife_aput(aout,'\\');
                mife_aput(aout,'n');
                break;
            case '\f':
                mife_aput(aout,'\\');
                mife_aput(aout,'f');
                break;
            case '\r':
                mife_aput(aout,'\\');
                mife_aput(aout,'r');
                break;
            case '\t':
                mife_aput(aout,'\\');
                mife_aput(aout,'t');
                break;
            case '\v':
                mife_aput(aout,'\\');
                mife_aput(aout,'v');
                break;
            case '\\':
                mife_aput(aout,'\\');
                mife_aput(aout,'\\');
                break;
            default:
                mife_aput(aout,c);
        }
}

/*
 * outputs integer
 */
static void
wisck_num_output(mife_aout * aout, uint32_t num) 
{
    uint32_t c;
    uint32_t div; 

    c = 0;
    for (div = 1000000000; div > 0; div /= 10) {
        c += (num / div) % 10;
        if (c) mife_aput(aout,'0' + (num / div) % 10);
    }
    if(!c) mife_aput(aout,'0');
}

static void
wisck_output(struct wisck_descriptor * wisck, struct wisck_rec *rec) 
{
    char ist;

    ist = 0;
    if (wisck->flags & WISCK_OLIN) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        wisck_num_output(wisck->aout, rec->line);
    }
    if (wisck->flags & WISCK_OREC) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        wisck_num_output(wisck->aout, rec->rec);
    }
    if (wisck->flags & WISCK_OOBJ) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        wisck_num_output(wisck->aout, rec->obj);
    }
    if (wisck->flags & WISCK_ONAM) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        if (rec->name) {
            wisck_escape_output(wisck->aout, rec->name);
        } else {
            mife_aput(wisck->aout,'\\');
            mife_aput(wisck->aout,'N');
        }
    }
    if (wisck->flags & WISCK_OVAL) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        if (rec->value) {
            wisck_escape_output(wisck->aout, rec->value);
        } else {
            mife_aput(wisck->aout,'\\');
            mife_aput(wisck->aout,'N');
        }
    }
    if (wisck->flags & WISCK_OCOM) {
        if (ist++ > 0) mife_aput(wisck->aout, '\t');
        if (rec->comm) {
            wisck_escape_output(wisck->aout, rec->comm);
        } else {
            mife_aput(wisck->aout,'\\');
            mife_aput(wisck->aout,'N');
        }
    }
    mife_aput(wisck->aout, '\n');
}

/*
 * Close wisck descriptor, releasing allocated memory
 */
int 
wisck_close(struct wisck_descriptor *wisck) 
{
    int result = 0;

    if (-1 == mife_close(wisck->mife)) {
        ifBLIN_QV2(wisck->flags)
            warn("mife_close failed");
        result = -1;
    }
    if (-1 == mife_aclose(wisck->aout)){
        ifBLIN_QV2(wisck->flags)
            warn("mife_aclose failed");
        result = -1;
    }
    free(wisck);
    return  result;
}
/*====================parse & output=======================*/

/*
 * Performs complete file converting. facade-function
 */
int 
wisck_convert(int infd, int outfd, uint32_t flags, unsigned char vsep, unsigned char csep) 
{
    struct wisck_descriptor *wisck;
    struct wisck_rec rec;

    if (NULL == (wisck = wisck_init(infd, outfd, flags, vsep, csep)))
        return -1; 
    while (1) {
        if (wisck_prepare(wisck, &rec)) {
            wisck_close(wisck);
            return -1;
        }
        switch (wisck_parse(wisck, &rec)) {
                case -1:
                        wisck_close(wisck);
                        return -1;
                        break;
                case 0:
                        if (-1 == wisck_close(wisck))
                        ifBLIN_QV2(wisck->flags) {
                            warn("parse complete, but wisck_close failed");
                            return -1;
                        }
                        return 0;
                        break;
                case 1:
                        wisck_output(wisck, &rec);
                        break;
        }
    }
    return 0; 
}
